#=========================================================================
# custom_exec.py
#=========================================================================
# The purpose of this custom_exec function is to turn dictionaries into
# ModuleDicts that results in much better JIT behavior under PyPy.
#
# Author : Shunning Jiang
# Date   : Feb 12, 2020

try:
  import __pypy__

  def _normalize_dict( _dict ):
    # Credit: Lin Cheng
    # return the original dictionary if not running on PyPy
    try:
      from __pypy__ import newdict, strategy
    except Exception:
      return _dict

    # return the original dictionary if already using ModuleDictStrategy
    if strategy( _dict ) == "ModuleDictStrategy":
      return _dict
    # create a new module dict
    new_dict = newdict("module")
    # copy over entries
    for key, value in _dict.items():
      new_dict[key] = value
    return new_dict

  def custom_exec( prog, _globals, _locals ):
    assert _globals is not None
    assert _locals  is not None

    norm_globals = _globals
    norm_locals  = _locals
    if _globals is not None:
      norm_globals = _normalize_dict( _globals )
    if _locals is not None:
      norm_locals  = _normalize_dict( _locals )

    exec( prog, norm_globals, norm_locals )
    # Local may have more stuff generated by the code.
    # We need to put this back to the original _locals
    if _locals is not None:
      _locals.update( norm_locals )

except Exception:
  custom_exec = exec



#=========================================================================
# TranslationImport_adhoc_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : Jun 5, 2019
"""Test ad-hoc components with SystemVerilog translation and import."""

import pytest

from pymtl3.datatypes import Bits16, Bits32, bitstruct, concat
from pymtl3.dsl import Component, InPort, Interface, OutPort, update
from pymtl3.passes.backends.verilog import VerilogPlaceholderPass
from pymtl3.passes.rtlir.util.test_utility import get_parameter
from pymtl3.stdlib.test_utils import TestVectorSimulator

from .. import VerilogTranslationImportPass
from ..testcases import (
    CaseChildExplicitModuleName,
    CaseMultiPlaceholderImport,
    CasePlaceholderTranslationRegIncr,
    CasePlaceholderTranslationVReg,
    CaseVIncludePopulation,
    CaseVLibsTranslation,
)
from ..translation.behavioral.test.VBehavioralTranslatorL1_test import (
    test_verilog_behavioral_L1,
)
from ..translation.behavioral.test.VBehavioralTranslatorL2_test import (
    test_verilog_behavioral_L2,
)
from ..translation.behavioral.test.VBehavioralTranslatorL3_test import (
    test_verilog_behavioral_L3,
)
from ..translation.behavioral.test.VBehavioralTranslatorL4_test import (
    test_verilog_behavioral_L4,
)
from ..translation.behavioral.test.VBehavioralTranslatorL5_test import (
    test_verilog_behavioral_L5,
)
from ..translation.structural.test.VStructuralTranslatorL1_test import (
    test_verilog_structural_L1,
)
from ..translation.structural.test.VStructuralTranslatorL2_test import (
    test_verilog_structural_L2,
)
from ..translation.structural.test.VStructuralTranslatorL3_test import (
    test_verilog_structural_L3,
)
from ..translation.structural.test.VStructuralTranslatorL4_test import (
    test_verilog_structural_L4,
)


def run_test( case ):
  _m = case.DUT()
  _m.elaborate()
  _m.set_metadata( VerilogTranslationImportPass.enable, True )
  _m.apply( VerilogPlaceholderPass() )
  m = VerilogTranslationImportPass()( _m )
  sim = TestVectorSimulator( m, case.TV, case.TV_IN, case.TV_OUT )
  sim.run_test()

@pytest.mark.parametrize(
  'case', get_parameter('case', test_verilog_behavioral_L1) + \
          get_parameter('case', test_verilog_behavioral_L2) + \
          get_parameter('case', test_verilog_behavioral_L3) + \
          get_parameter('case', test_verilog_behavioral_L4) + \
          get_parameter('case', test_verilog_behavioral_L5) + \
          get_parameter('case', test_verilog_structural_L1) + \
          get_parameter('case', test_verilog_structural_L2) + \
          get_parameter('case', test_verilog_structural_L3) + \
          get_parameter('case', test_verilog_structural_L4) + [
            CasePlaceholderTranslationVReg,
            CasePlaceholderTranslationRegIncr,
            CaseVIncludePopulation,
            CaseVLibsTranslation,
            CaseChildExplicitModuleName,
          ]
)
def test_verilog_translation_import_adhoc( case ):
  run_test( case )

def test_pymtl_top_multi_placeholder():
  case = CaseMultiPlaceholderImport
  m = case.DUT()
  m.elaborate()
  m.apply( VerilogPlaceholderPass() )
  m = VerilogTranslationImportPass()( m )
  sim = TestVectorSimulator( m, case.TV, case.TV_IN, case.TV_OUT )
  sim.run_test()

def test_bitstruct_same_name_different_fields():
  class A:
    @bitstruct
    class A:
      a: Bits32

  class B:
    @bitstruct
    class A:
      a: Bits16

  class InIfc( Interface ):
    def construct( s, Type ):
      s.in_ = InPort( Type )

  class OutIfc( Interface ):
    def construct( s, Type ):
      s.out = OutPort( Type )

  class DUT( Component ):
    def construct( s ):
      s.in1 = InIfc(A.A)
      s.in2 = InIfc(B.A)
      s.out2 = OutPort(Bits16)

      s.out2 //= s.in2.in_.a

  class TMP( Component ):
    def construct( s ):
      s.out = OutPort(B.A)
      @update
      def drive():
        s.out @= 0

  class Top( Component ):
    def construct( s ):
      s.dut = DUT()
      s.tmp = TMP()
      s.tmp.out //= s.dut.in2.in_

  m = Top()
  m.elaborate()
  m.dut.set_metadata( VerilogTranslationImportPass.enable, True )
  m.apply( VerilogPlaceholderPass() )
  m = VerilogTranslationImportPass()( m )
  m.dut.finalize()



#=========================================================================
# TranslationImport_stdlib_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : Jun 15, 2019
"""Test stdlib RTL components with SystemVerilog translation and import.

We reuse all the test cases from stdlib test files. To achieve this I
overwrite the original reference to the test function used in stdlib test
files and add my own test function (the `run_test`s). However we need
to make sure the orignal reference is not lost and is restored after
finishing each test (no matter it fails or passes).
"""

from pymtl3.passes.PassGroups import DefaultPassGroup

from pymtl3.passes.backends.verilog.test.TranslationImport_stdlib_test import (
    test_arbiter_rr_arb_4,
    test_arbiter_rr_arb_en_4,
    test_crossbar3,
    test_encoder_5_directed,
)
from pymtl3.passes.rtlir.util.test_utility import do_test
from pymtl3.stdlib.test_utils import TestVectorSimulator

from pymtl3.stdlib.stream.test.queues_test import test_normal1_simple as _n1
from pymtl3.stdlib.stream.test.queues_test import test_normal2_simple as _n2
from pymtl3.stdlib.stream.test.queues_test import test_bypass1_simple as _b1
from pymtl3.stdlib.stream.test.queues_test import test_pipe1_simple as _p1

from ..YosysTranslationImportPass import YosysTranslationImportPass


def local_do_test( _m ):
  _m.elaborate()
  # Mark component `_m` as to be translated and imported
  _m.set_metadata( YosysTranslationImportPass.enable, True )
  m = YosysTranslationImportPass()( _m )
  sim = TestVectorSimulator( m, _m._tvs, _m._tv_in, _m._tv_out )
  sim.run_test()

#-------------------------------------------------------------------------
# Queue tests
#-------------------------------------------------------------------------

# Swap run_test_queue with modified version that calls the local run_sim
def _run_queue_test( test_func ):

  def _run_sim( th, cmdline_opts, duts ):
    th.elaborate()
    dut_objs = []

    for dut in duts:
      dut_objs.append( eval(f'th.{dut}') )

    for obj in dut_objs:
      obj.set_metadata( YosysTranslationImportPass.enable, True )

    th = YosysTranslationImportPass()( th )

    th.apply( DefaultPassGroup() )
    th.sim_reset()

    while not th.done() and th.sim_cycle_count() < 100:
      th.sim_tick()

    assert th.sim_cycle_count() < 100

  original_run_sim = test_func.__globals__['run_sim']
  test_func.__globals__['run_sim'] = _run_sim
  try:
    test_func( None )
  finally:
    test_func.__globals__['run_sim'] = original_run_sim

def test_normal1_simple():
  _run_queue_test( _n1 )

def test_normal2_simple():
  _run_queue_test( _n2 )

def test_bypass1_simple():
  _run_queue_test( _b1 )

def test_pipe1_simple():
  _run_queue_test( _p1 )



#=========================================================================
# ImportedObject_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : Jun 2, 2019
"""Test if the imported object works correctly."""

from pymtl3.passes.backends.verilog import VerilogPlaceholderPass
from pymtl3.passes.backends.verilog.import_.test.ImportedObject_test import (
    test_adder,
    test_normal_queue_implicit_top_module,
    test_normal_queue_interface,
    test_normal_queue_params,
    test_reg,
    test_reg_external_trace,
    test_reg_incomplete_portmap,
    test_reg_infer_external_trace,
    test_vl_uninit,
)
from pymtl3.passes.backends.yosys import YosysTranslationImportPass
from pymtl3.passes.rtlir.util.test_utility import do_test
from pymtl3.stdlib.test_utils import TestVectorSimulator


def local_do_test( _m ):
  _m.elaborate()
  if not hasattr( _m, "_no_trans_import" ):
    _m.set_metadata( YosysTranslationImportPass.enable, True )
  _m.apply( VerilogPlaceholderPass() )
  m = YosysTranslationImportPass()( _m )
  sim = TestVectorSimulator( m, _m._tvs, _m._tv_in, _m._tv_out )
  sim.run_test()



#=========================================================================
# utility.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 27, 2019
"""Provide helper methods that might be useful to verilog passes."""

from pymtl3.passes.rtlir import RTLIRDataType as rdt
from pymtl3.passes.rtlir import RTLIRGetter
from pymtl3.passes.rtlir import RTLIRType as rt

#-----------------------------------------------------------------------
# gen_packed_ports
#-----------------------------------------------------------------------

def get_rtype( _rtype ):
  if isinstance( _rtype, rt.Array ):
    n_dim = _rtype.get_dim_sizes()
    rtype = _rtype.get_sub_type()
  else:
    n_dim = []
    rtype = _rtype
  return n_dim, rtype

#-----------------------------------------------------------------------
# gen_mapped_ports
#-----------------------------------------------------------------------

def gen_mapped_ports( m, port_map, has_clk=True, has_reset=True, sep='__' ):
  """Return a list of (pname, vname, rt.Port/rt.Array ) that has all ports
  of `rtype`. This method performs SystemVerilog backend-specific name
  mangling and returns all ports that appear in the interface of component
  `rtype`. Each tuple contains a port or an array of port that has any data type
  allowed in RTLIRDataType.
  Shunning: Now we also take port_map into account. Two points to note:
  1. If a port's pname appears as a key in port_map, we need to use the
     corresponding value as vname
  2. For an n-D array of ports, we enforce the rule that assumes either no
     element is mapped in port_map, or _all_ of the elements are mapped.
  """

  port_map = {p._dsl._my_name:n for p, n in port_map.items()}

  def _mangle_vector( pname, vname, port, dtype, port_idx ):
    return [([pname], port_map[pname] if pname in port_map else vname,
             rt.Port(port.direction, dtype), port_idx)]

  def _mangle_struct( pname, vname, port, dtype, port_idx ):
    ret = []
    for field_name, field_dtype in dtype.get_all_properties().items():
      ret += _mangle_dtype(f"{pname}.{field_name}", f"{vname}{sep}{field_name}",
                           port, field_dtype, port_idx)
    return ret

  def _mangle_packed( pname, vname, port, dtype, n_dim, port_idx ):
    if not n_dim:
      return _mangle_dtype( pname, vname, port, dtype, port_idx )
    else:
      ret = []
      for i in range(n_dim[0]):
        ret += _mangle_packed( f"{pname}[{i}]", f"{vname}{sep}{i}", port, dtype, n_dim[1:], port_idx )
      return ret

  def _mangle_dtype( pname, vname, port, dtype, port_idx ):
    if isinstance(dtype, rdt.Vector):
      return _mangle_vector(pname, vname, port, dtype, port_idx)
    elif isinstance(dtype, rdt.Struct):
      return _mangle_struct(pname, vname, port, dtype, port_idx)
    elif isinstance(dtype, rdt.PackedArray):
      return _mangle_packed(pname, vname, port,
                            dtype.get_sub_dtype(), dtype.get_dim_sizes(), port_idx)
    else:
      assert False, f'unrecognized data type {dtype}!'

  def _mangle_port( pname, vname, port, n_dim, port_idx ):
    # Normal port
    if not n_dim:
      return _mangle_dtype( pname, vname, port, port.get_dtype(), port_idx )

    # Handle port array. We just assume if one element of the port array
    # is mapped, we need the user to map every element in the array.
    else:
      ret = []
      for i in range(n_dim[0]):
        ret += _mangle_port(f"{pname}[{i}]", f"{vname}{sep}{i}", port, n_dim[1:], port_idx)
      return ret

  def _mangle_ifc( pname, vname, ifc, n_dim, port_idx ):
    ret = []
    if not n_dim:
      for port_name, port_rtype in ifc.get_all_properties_packed():
        port_n_dim, port_rtype = get_rtype( port_rtype )
        if isinstance(port_rtype, rt.InterfaceView):
          ret += _mangle_ifc(f"{pname}.{port_name}", f"{vname}{sep}{port_name}",
                              port_rtype, port_n_dim, port_idx)
        elif isinstance(port_rtype, rt.Port):
          ret += _mangle_port(f"{pname}.{port_name}", f"{vname}{sep}{port_name}",
                              port_rtype, port_n_dim, port_idx)
        else:
          assert False, "unrecognized interface/port {port_rtype}!"
    else:
      for i in range(n_dim[0]):
        ret += _mangle_ifc(f"{pname}[{i}]", f"{vname}{sep}{i}", ifc, n_dim[1:], port_idx+1)
    return ret

  # We start from all packed ports/interfaces, and unpack arrays if
  # it is found in a port.
  rtype = RTLIRGetter(cache=False).get_component_ifc_rtlir(m)
  ret = []

  for name, port in rtype.get_ports_packed():
    if not has_clk and name == 'clk':      continue
    if not has_reset and name == 'reset':  continue
    p_n_dim, p_rtype = get_rtype( port )
    ret += _mangle_port( name, name, p_rtype, p_n_dim, 0 )

  for name, ifc in rtype.get_ifc_views_packed():
    i_n_dim, i_rtype = get_rtype( ifc )
    ret += _mangle_ifc( name, name, i_rtype, i_n_dim, 0 )

  return ret



"""
======================================================================
arbiters.py
======================================================================
Collection of arbiter implementations based on vc-Arbiters.
"""

from pymtl3 import *

from .registers import RegEnRst


#-----------------------------------------------------------------------
# RoundRobinArbiter
#-----------------------------------------------------------------------
class RoundRobinArbiter( Component ):

  def construct( s, nreqs ):

    nreqsX2 = nreqs * 2
    Type    = mk_bits( nreqs )

    s.reqs   = InPort ( Type )
    s.grants = OutPort( Type )

    # priority enable

    s.priority_en = Wire()

    # priority register

    s.priority_reg = m = RegEnRst( Type, reset_value = 1 )

    connect( m.en,           s.priority_en )
    connect( m.in_[1:nreqs], s.grants[0:nreqs-1] )
    connect( m.in_[0],       s.grants[nreqs-1] )

    s.kills        = Wire( 2*nreqs + 1 )
    s.priority_int = Wire( 2*nreqs )
    s.reqs_int     = Wire( 2*nreqs )
    s.grants_int   = Wire( 2*nreqs )

    @update
    def comb_reqs_int():
      s.reqs_int [    0:nreqs  ] @= s.reqs
      s.reqs_int [nreqs:nreqsX2] @= s.reqs

    @update
    def comb_grants():
      for i in range( nreqs ):
        s.grants[i] @= s.grants_int[i] | s.grants_int[nreqs+i]

    @update
    def comb_priority_en():
      s.priority_en @= s.grants != 0

    @update
    def comb_priority_int():
      s.priority_int[    0:nreqs  ] @= s.priority_reg.out
      s.priority_int[nreqs:nreqsX2] @= 0

    @update
    def comb_kills():
      s.kills[0] @= 1
      for i in range( nreqsX2 ):
        if s.priority_int[i]:
          s.kills[i+1] @= s.reqs_int[i]
        else:
          s.kills[i+1] @= s.kills[i] | ( ~s.kills[i] & s.reqs_int[i] )

    @update
    def comb_grants_int():
      for i in range( nreqsX2 ):
        if s.priority_int[i]:
          s.grants_int[i] @= s.reqs_int[i]
        else:
          s.grants_int[i] @= ~s.kills[i] & s.reqs_int[i]

  def line_trace( s ):
    return f"{s.reqs} | {s.grants}"

#-----------------------------------------------------------------------
# RoundRobinArbiterEn
#-----------------------------------------------------------------------
class RoundRobinArbiterEn( Component ):
  '''Round Robin Arbiter model with an enable bit.

  Only when 1) the arbiter enable bit is set high AND 2) there is some
  valid arbitration, the priority register is updated. This breaks the
  val-rdy dependency.
  '''
  def construct( s, nreqs ):

    nreqsX2 = nreqs * 2
    Type    = mk_bits( nreqs )

    s.en     = InPort ()
    s.reqs   = InPort ( Type )
    s.grants = OutPort( Type )

    s.priority_en = Wire()

    s.priority_reg = m = RegEnRst( mk_bits( nreqs ), reset_value = 1 )
    m.en           //= s.priority_en
    m.in_[1:nreqs] //= s.grants[0:nreqs-1]
    m.in_[0]       //= s.grants[nreqs-1]

    s.kills        = Wire( 2*nreqs + 1 )
    s.priority_int = Wire( 2*nreqs )
    s.reqs_int     = Wire( 2*nreqs )
    s.grants_int   = Wire( 2*nreqs )

    @update
    def comb_reqs_int():
      s.reqs_int [    0:nreqs  ] @= s.reqs
      s.reqs_int [nreqs:nreqsX2] @= s.reqs

    @update
    def comb_grants():
      for i in range( nreqs ):
        s.grants[i] @= s.grants_int[i] | s.grants_int[nreqs+i]

    @update
    def comb_priority_en():
      s.priority_en @= ( s.grants != 0 ) & s.en

    @update
    def comb_priority_int():
      s.priority_int[    0:nreqs  ] @= s.priority_reg.out
      s.priority_int[nreqs:nreqsX2] @= 0

    @update
    def comb_kills():
      s.kills[0] @= 1
      for i in range( nreqsX2 ):
        if s.priority_int[i]:
          s.kills[i+1] @= s.reqs_int[i]
        else:
          s.kills[i+1] @= s.kills[i] | ( ~s.kills[i] & s.reqs_int[i] )

    @update
    def comb_grants_int():
      for i in range( nreqsX2 ):
        if s.priority_int[i]:
          s.grants_int[i] @= s.reqs_int[i]
        else:
          s.grants_int[i] @= ~s.kills[i] & s.reqs_int[i]

  def line_trace( s ):
    return f"{s.reqs} | {s.grants}"



"""
========================================================================
SourceRTL
========================================================================
Test sources with RTL interfaces.

Author : Shunning Jiang
  Date : Feb 12, 2021
"""
from collections import deque
from copy import deepcopy

from pymtl3 import *
from .ifcs import SendIfcRTL


class SourceRTL( Component ):

  def construct( s, Type, msgs, initial_delay=0, interval_delay=0 ):

    # Interface

    s.send = SendIfcRTL( Type )

    # Data

    s.msgs = deepcopy(msgs)

    # TODO: use wires and ROM to make it translatable
    s.idx = 0
    s.count = 0

    @update_ff
    def up_src():
      if s.reset:
        s.idx   = 0
        s.count = initial_delay
        s.send.val <<= 0

      else:
        if s.send.val & s.send.rdy:
          s.idx += 1
          s.count = interval_delay

        if s.count > 0:
          s.count -= 1
          s.send.val <<= 0

        else: # s.count == 0
          if s.idx < len(s.msgs):
            s.send.val <<= 1
            s.send.msg <<= s.msgs[s.idx]
          else:
            s.send.val <<= 0


  def done( s ):
    return s.idx >= len(s.msgs)

  # Line trace

  def line_trace( s ):
    return f"{s.send}"



#=========================================================================
# utility.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 27, 2019
"""Provide helper methods that might be useful to verilog passes."""

import os
import shutil
import textwrap
from collections import deque
from hashlib import blake2b

from pymtl3.passes.rtlir import RTLIRDataType as rdt
from pymtl3.passes.rtlir import RTLIRGetter
from pymtl3.passes.rtlir import RTLIRType as rt
from pymtl3.passes.rtlir.util.utility import get_component_full_name


def make_indent( src, nindent ):
  """Add nindent indention to every line in src."""
  indent = '  '
  for idx, s in enumerate( src ):
    src[ idx ] = nindent * indent + s

def get_component_unique_name( c_rtype ):
  full_name = get_component_full_name( c_rtype )
  special_chars = [' ', '<', '>', '.', '[', ']']

  if len( full_name ) < 64 and not any([c in full_name for c in special_chars]):
    return full_name

  comp_name = c_rtype.get_name()
  param_hash = blake2b(digest_size = 8)
  param_hash.update(full_name[len(comp_name):].encode('ascii'))
  param_name = param_hash.hexdigest()
  return comp_name + "__" + param_name

def wrap( s ):
  col = shutil.get_terminal_size().columns
  return "\n".join(sum((textwrap.wrap(line, col) for line in s.split("\n")), []))

def expand( v ):
  return os.path.expanduser(os.path.expandvars(v))

def pretty_concat( *strings ):
  ret = ' '.join([s for s in strings[:-1] if s])
  if strings[-1] == ';':
    ret += ';'
  else:
    ret += f' {strings[-1]}'
  return ret

def get_dir( cur_file ):
  return os.path.dirname(os.path.abspath(cur_file))+os.path.sep

def get_file_hash( file_path ):
  with open(file_path) as fd:
    hash_inst = blake2b()
    string = ''.join( fd.readlines() ).encode( 'ascii' )
    hash_inst.update(string)
    return hash_inst.hexdigest()

def get_lean_verilog_file( file_path ):
  with open(file_path) as fd:
    file_v = [x for x in fd.readlines() if x != '\n' and not x.startswith('//')]
  return file_v

def get_hash_of_lean_verilog( file_path ):
  hash_inst = blake2b()
  v = get_lean_verilog_file( file_path )
  string = ''.join(v).encode( 'ascii' )
  hash_inst.update(string)
  return hash_inst.hexdigest()

def verilog_cmp( tmp, out ):
  tmp_v, out_v = get_lean_verilog_file(tmp), get_lean_verilog_file(out)
  is_same_len = len(tmp_v) == len(out_v)
  if is_same_len:
    for i in range(len(out_v)):
      if out_v[i] != tmp_v[i]:
        return False
  return is_same_len

verilog_keyword = [
  # Verilog-1995 reserved keywords
  "always", "and", "assign", "begin", "buf", "bufif0", "bufif1", "case",
  "casex", "casez", "cmos", "deassign", "default", "defparam", "disable",
  "edge", "else", "end", "endcase", "endmodule", "endfunction", "endprimitive",
  "endspecify", "endtable", "endtask", "event", "for", "force", "forever",
  "fork", "function", "highz0", "highz1", "if", "ifnone", "initial",
  "inout", "input", "output", "integer", "join", "large", "macromodule",
  "medium", "module", "nand", "negedge", "nmos", "nor", "not", "notif0",
  "notif1", "or", "output", "parameter", "pmos", "posedge", "primitive",
  "pull0", "pull1", "pullup", "pulldown", "rcmos", "real", "realtime",
  "reg", "release", "repeat", "rnmos", "rpmos", "rtran", "rtranif0",
  "rtranif1", "scalared", "small", "specify", "specparam", "strong0",
  "strong1", "supply0", "supply1", "table", "task", "time", "tran",
  "tranif0", "tranif1", "tri", "tri0", "tri1", "triand", "trior",
  "trireg", "vectored", "wait", "wand", "weak0", "weak1", "while",
  "wire", "wor", "xnor", "xor",
  # Verilog-2001 reserved keywords
  "automatic", "cell", "config", "design", "endconfig", "endgenerate",
  "generate", "genvar", "incdir", "include", "instance", "liblist",
  "library", "localparam", "noshowcancelled", "pulsestyle_onevent",
  "pulsestyle_ondetect", "showcancelled", "signed", "unsigned", "use",
  # Verilog-2005 reserved keywords
  "uwire",
  # SystemVerilog-2005 reserved keywords
  "alias", "always_comb", "always_ff", "always_latch", "assert", "assume",
  "before", "bind", "bins", "binsof", "bit", "break", "byte", "chandle",
  "class", "clocking", "const", "constraint", "context", "continue",
  "cover", "covergroup", "coverpoint", "cross", "dist", "do", "endclass",
  "endclocking", "endgroup", "endinterface", "endpackage", "endprimitive",
  "endprogram", "endproperty", "endsequence", "enum", "expect", "export",
  "extends", "extern", "final", "first_match", "foreach", "forkjoin",
  "iff", "ignore_bins", "illegal_bins", "import", "inside", "int", "interface",
  "intersect", "join_any", "join_none", "local", "logic", "longint", "matches",
  "modport", "new", "null", "package", "packed", "priority", "program",
  "property", "protected", "pure", "rand", "randc", "randcase", "randsequence",
  "ref", "return", "sequence", "shortint", "shortreal", "solve", "static",
  "string", "struct", "super", "tagged", "this", "throughout", "timeprecision",
  "timeunit", "type", "typedef", "union", "unique", "var", "virtual", "void",
  "wait_order", "wildcard", "with", "within"
]

verilog_reserved = set( verilog_keyword )

#-----------------------------------------------------------------------
# get_rtype
#-----------------------------------------------------------------------

def get_rtype( _rtype ):
  if isinstance( _rtype, rt.Array ):
    n_dim = _rtype.get_dim_sizes()
    rtype = _rtype.get_sub_type()
  else:
    n_dim = []
    rtype = _rtype
  return n_dim, rtype

#-----------------------------------------------------------------------
# gen_mapped_ports
#-----------------------------------------------------------------------

def gen_mapped_ports( m, raw_port_map, has_clk=True, has_reset=True, sep='__' ):
  """Return a list of (pname, vname, rt.Port/rt.Array ) that has all ports
  of `rtype`. This method performs SystemVerilog backend-specific name
  mangling and returns all ports that appear in the interface of component
  `rtype`. Each tuple contains a port or an array of port that has any data type
  allowed in RTLIRDataType.
  Shunning: Now we also take port_map into account. Two points to note:
  1. If a port's pname appears as a key in port_map, we need to use the
     corresponding value as vname
  2. For an n-D array of ports, we enforce the rule that assumes either no
     element is mapped in port_map, or _all_ of the elements are mapped.
  """

  # Shunning: this handles the case where the port is in interface since
  # we need to actually extract 'minion.req.en' from 's.x.y.minion.req.en'
  port_map = {}
  l = len(repr(m))
  for p, mapped_name in raw_port_map.items():
    assert p.get_host_component() is m
    port_map[ repr(p)[l+1:] ] = mapped_name

  # [pnames], vname, rtype, port_idx

  def _mangle_port( pname, vname, port, n_dim ):

    # Normal port
    if not n_dim:
      return [ ( [pname], port_map[pname] if pname in port_map else vname, port, 0 ) ]

    # Handle port array. We just assume if one element of the port array
    # is mapped, we need the user to map every element in the array.
    found = tot = 0
    all_ports = []
    Q = deque( [ (pname, vname, port, n_dim ) ] )
    while Q:
      _pname, _vname, _port, _n_dim = Q.popleft()
      if not _n_dim:
        if _pname in port_map:
          found += 1
          _vname = port_map[_pname]
        all_ports.append( ( [_pname], _vname, _port, 0 ) )
      else:
        for i in range( _n_dim[0] ):
          Q.append( (f"{_pname}[{i}]", f"{_vname}{sep}{i}", _port, _n_dim[1:]) )

    assert found == len(all_ports) or found == 0, \
        f"{pname} is an {len(n_dim)}-D array of ports with {len(all_ports)} ports in total, " \
        f" but only {found} of them is mapped. Please either map all of them or none of them."

    if not found:
      return [ ( [pname], vname, rt.Array( n_dim, port ), 0 ) ]
    else:
      return all_ports

  def _is_ifc_mapped( pname, vname, rtype, n_dim ):
    found, tot, flatten_ports = 0, 0, []
    # pname, vname, rtype, n_dim
    Q = deque( [ (pname, vname, rtype, n_dim, 0) ] )

    while Q:
      _pname, _vname, _rtype, _n_dim, _prev_dims = Q.popleft()
      if _n_dim:
        for i in range(_n_dim[0]):
          Q.append((f"{_pname}[{i}]", f"{_vname}{sep}{i}", _rtype, _n_dim[1:], _prev_dims+1))
      else:
        if isinstance( _rtype, rt.Port ):
          # Port inside the interface
          tot += 1
          if _pname in port_map:
            found += 1
            flatten_ports.append(([_pname], port_map[_pname], _rtype, _prev_dims))
          else:
            flatten_ports.append(([_pname], _vname, _rtype, _prev_dims))
        elif isinstance( _rtype, rt.InterfaceView ):
          # Interface (nested)
          for sub_name, sub_rtype in _rtype.get_all_properties_packed():
            sub_n_dim, sub_rtype = get_rtype( sub_rtype )
            Q.append((f"{_pname}.{sub_name}", f"{_vname}{sep}{sub_name}", sub_rtype, sub_n_dim, _prev_dims))
        else:
          assert False, f"{_pname} is not interface(s) or port(s)!"

    assert (found == 0) or (found == tot), \
        f"{name} is an interface that has {tot} ports in total, " \
        f" but only {found} of them is mapped. Please either map all of them or none of them."
    return (found == tot), flatten_ports

  def _gen_packed_ifc( pname, vname, ifc, n_dim ):
    packed_ifc, ret = [], []
    Q = deque( [ (pname, vname, ifc, n_dim, [], 0) ] )
    while Q:
      _pname, _vname, _rtype, _n_dim, _prev_n_dim, _port_idx = Q.popleft()

      if isinstance( _rtype, rt.Port ):
        if not (_prev_n_dim + _n_dim):
          new_rtype = _rtype
        else:
          new_rtype = rt.Array(_prev_n_dim+_n_dim, _rtype)
        packed_ifc.append((_pname, _vname, new_rtype, _port_idx))

      elif isinstance( _rtype, rt.InterfaceView ):
        if _n_dim:
          new_prev_n_dim = _prev_n_dim + [_n_dim[0]]
          for i in range(_n_dim[0]):
            Q.append((f"{_pname}[{i}]", _vname, _rtype, _n_dim[1:], new_prev_n_dim, _port_idx+1))
        else:
          new_prev_n_dim = _prev_n_dim
          for sub_name, sub_rtype in _rtype.get_all_properties_packed():
            sub_n_dim, sub_rtype = get_rtype( sub_rtype )
            Q.append((f"{_pname}.{sub_name}", f"{_vname}{sep}{sub_name}",
                      sub_rtype, sub_n_dim, new_prev_n_dim, _port_idx))
      else:
        assert False, f"{_pname} is not interface(s) or port(s)!"

    # Merge entries whose vnames are the same. The result will have a list for
    # the pnames.
    names = set()
    for _, vname, rtype, port_idx in packed_ifc:
      if vname not in names:
        names.add( vname )
        ret.append(([], vname, rtype, port_idx))
        for _pname, _vname, _, _port_idx in packed_ifc:
          if vname == _vname:
            assert _port_idx == port_idx
            ret[-1][0].append(_pname)

    return ret

  def _mangle_ifc( pname, vname, ifc, n_dim ):
    is_mapped, flatten_ifc = _is_ifc_mapped( pname, vname, ifc, n_dim )
    if is_mapped:
      return flatten_ifc
    else:
      return _gen_packed_ifc( pname, vname, ifc, n_dim )

  # We start from all packed ports/interfaces, and unpack arrays if
  # it is found in a port.
  rtype = RTLIRGetter(cache=False).get_component_ifc_rtlir(m)
  ret = []

  for name, port in rtype.get_ports_packed():
    if not has_clk and name == 'clk':      continue
    if not has_reset and name == 'reset':  continue
    p_n_dim, p_rtype = get_rtype( port )
    ret += _mangle_port( name, name, p_rtype, p_n_dim )

  for name, ifc in rtype.get_ifc_views_packed():
    i_n_dim, i_rtype = get_rtype( ifc )
    ret += _mangle_ifc( name, name, i_rtype, i_n_dim )

  return ret



from .SinkRTL import SinkRTL
from .SourceRTL import SourceRTL
from . import ifcs
from .queue_adapters import RecvQueueAdapter, SendQueueAdapter
from .queues import NormalQueueRTL, PipeQueueRTL, BypassQueueRTL
from .magic_memory import MagicMemoryRTL
from . import fl



#=========================================================================
# PlaceholderConfigs.py
#=========================================================================
# Author : Peitian Pan
# Date   : Jul 28, 2019
"""Configuration of Placeholders."""

import os

from pymtl3.passes.PassConfigs import BasePassConfigs, Checker
from pymtl3.passes.PlaceholderPass import PlaceholderPass


def expand( path ):
  return os.path.expanduser(os.path.expandvars(path))

class PlaceholderConfigs( BasePassConfigs ):

  Options = {
    # Should placeholder passes handle this instance?
    "enable" : True,

    # Does the module to be imported has `clk` port?
    "has_clk" : True,

    # Does the module to be imported has `reset` port?
    "has_reset" : True,
  }

  Checkers = {
    ("enable", "has_clk", "has_reset") :
      Checker( lambda v: isinstance(v, bool), "expects a boolean" ),
  }

  Pass = PlaceholderPass



from .MagicMemoryCL import MagicMemoryCL
from .MagicMemoryFL import MagicMemoryFL
from .mem_ifcs import (
    MemMasterIfcCL,
    MemMasterIfcFL,
    MemMasterIfcRTL,
    MemMinionIfcCL,
    MemMinionIfcFL,
    MemMinionIfcRTL,
)
from .MemMsg import MemMsgType, mk_mem_msg, mk_mem_req_msg, mk_mem_resp_msg
from .ROMRTL import CombinationalROMRTL, SequentialROMRTL



"""
========================================================================
ROMRTL.py
========================================================================
Multiported ROM

Author : Shunning Jiang
Date   : June 18, 2020
"""

from pymtl3 import *


class CombinationalROMRTL( Component ):

  def construct( s, Type, num_entries, data, num_ports=1 ):
    assert len(data) == num_entries

    s.raddr = [ InPort( clog2(num_entries) ) for _ in range(num_ports) ]
    s.rdata = [ OutPort( Type )              for _ in range(num_ports) ]

    s.mem = [ Wire(Type) for _ in range(num_entries) ]
    for i in range(num_entries):
      s.mem[i] //= data[i]

    @update
    def up_read_rom():
      for i in range(num_ports):
        s.rdata[i] @= s.mem[ s.raddr[i] ]

class SequentialROMRTL( Component ):

  def construct( s, Type, num_entries, data, num_ports=1 ):
    assert len(data) == num_entries

    s.raddr = [ InPort( clog2(num_entries) ) for _ in range(num_ports) ]
    s.rdata = [ OutPort( Type )              for _ in range(num_ports) ]

    s.mem = [ Wire(Type) for _ in range(num_entries) ]
    for i in range(num_entries):
      s.mem[i] //= data[i]

    @update_ff
    def up_read_rom():
      for i in range(num_ports):
        s.rdata[i] <<= s.mem[ s.raddr[i] ]
  #-----------------------------------------------------------------------
  # line_trace
  #-----------------------------------------------------------------------

  def line_trace( s ):
    return "|".join( [ f"[{s.raddr[i]}]->{s.rdata[i]}" for i in range(len(s.raddr)) ] )



#=========================================================================
# StructuralRTLIRGenL1Pass.py
#=========================================================================
# Author : Shunning Jiang, Peitian Pan
# Date   : Apr 3, 2019
"""Provide L1 structural RTLIR generation pass."""

from pymtl3 import MetadataKey
from pymtl3.passes.PlaceholderConfigs import PlaceholderConfigs
# from pymtl3.passes.rtlir.RTLIRPass import RTLIRPass
from pymtl3.passes.rtlir.errors import RTLIRConversionError
from pymtl3.passes.rtlir.rtype.RTLIRType import RTLIRGetter

from .StructuralRTLIRGenL0Pass import StructuralRTLIRGenL0Pass
from .StructuralRTLIRSignalExpr import gen_signal_expr


class StructuralRTLIRGenL1Pass( StructuralRTLIRGenL0Pass ):

  def __init__( s, inst_conns ):
    s.inst_conns = inst_conns

  def __call__( s, tr_top ):
    """ generate structural RTLIR for component `tr_top` """
    c = s.__class__
    s.tr_top = tr_top
    if not tr_top.has_metadata( c.rtlir_getter ):
      tr_top.set_metadata( c.rtlir_getter, RTLIRGetter(cache=True) )

    try:
      s._gen_metadata( tr_top )
    except AssertionError as e:
      msg = '' if not e.args is None else e.args[0]
      raise RTLIRConversionError( tr_top, msg )

  def _gen_metadata( s, m ):
    c = s.__class__

    # Generate RTLIR types
    rtlir_type = s.tr_top.get_metadata( c.rtlir_getter ).get_rtlir( m )
    m.set_metadata( c.rtlir_type, rtlir_type )

    # Generate constants
    consts = []
    rtype = rtlir_type
    const_types = rtype.get_consts_packed()
    for const_name, const_rtype in const_types:
      assert hasattr(m, const_name), \
        f"Internal error: {const_name} is not a member of {m}"
      const_instance = getattr(m, const_name)
      consts.append( ( const_name, const_rtype, const_instance ) )

    m.set_metadata( c.consts, consts )

    # Sort connections
    m_conns_set   = s.inst_conns[m]
    ordered_conns = [ *m.get_connect_order() ]

    # NOTE: this assertion can fail due to connections that
    # are made outside the component that has them. so i'm removing
    # this for now until we can figure out a better way to do sanity
    # check here.
    # assert len(ordered_conns) == len(m_conns_set)

    for i, x in enumerate(ordered_conns):
      if x not in m_conns_set:
        x = (x[1], x[0])
        assert x in m_conns_set, "There is a connection missing from "\
                                 "connect_order. Please contact PyMTL developers!"
        ordered_conns[i] = x

    connections = [ (gen_signal_expr(m, x[0]), gen_signal_expr(m, x[1])) for x in ordered_conns ]

    m.set_metadata( c.connections, connections )



import greenlet
from pymtl3 import *
from pymtl3.extra import clone_deepcopy
from pymtl3.stdlib.mem.MemMsg import MemMsgType
from .ifcs import RecvIfcRTL, SendIfcRTL, MasterIfcRTL

class RecvQueueAdapter( Component ):

  @blocking
  def deq( s ):
    while s.entry is None:
      greenlet.getcurrent().parent.switch(0)
    ret = s.entry
    s.entry = None
    return ret

  def construct( s, Type ):
    s.recv = RecvIfcRTL( Type )
    s.entry = None

    @update_once
    def up_recv_rdy():
      s.recv.rdy @= (s.entry is None)

    @update_once
    def up_recv_msg():
      if (s.entry is None) & s.recv.val:
        s.entry = clone_deepcopy( s.recv.msg )

    s.add_constraints( M( s.deq )       < U( up_recv_rdy ), # deq before recv in a cycle -- pipe behavior
                       U( up_recv_rdy ) < U( up_recv_msg ) )

  def line_trace( s ):
    return "{}(){}".format( s.recv, s.send )

class SendQueueAdapter( Component ):

  @blocking
  def enq( s, msg ):
    while s.entry is not None:
      greenlet.getcurrent().parent.switch(0)

    s.entry = clone_deepcopy(msg)

  def construct( s, Type ):
    s.send = SendIfcRTL( Type )
    s.entry = None

    s.sent = Wire()

    @update
    def up_send():
      if s.entry is None:
        s.send.val @= 0
      else:
        s.send.val @= 1
        s.send.msg @= s.entry

    @update_ff
    def up_sent():
      s.sent <<= s.send.val & s.send.rdy

    @update_once
    def up_clear():
      if s.sent: # constraints reverse this
        s.entry = None

    s.add_constraints( U( up_clear ) < M( s.enq ),
                       M( s.enq )    < U( up_send ) )

  def line_trace( s ):
    return "{}(){}".format( s.recv, s.send )

class MemMasterAdapter( Component ):

  @blocking
  def read( s, addr, nbytes ):
    while s.req_entry is not None:
      greenlet.getcurrent().parent.switch(0)

    s.req_entry = s.create_req( MemMsgType.READ, 0, addr, nbytes )

    while s.resp_entry is None:
      greenlet.getcurrent().parent.switch(0)

    ret = s.resp_entry.data[0:nbytes<<3]
    s.resp_entry = None
    return ret

  @blocking
  def write( s, addr, nbytes, data ):
    while s.req_entry is not None:
      greenlet.getcurrent().parent.switch(0)

    s.req_entry = s.create_req( MemMsgType.WRITE, 0, addr, nbytes, data )

    while s.resp_entry is None:
      greenlet.getcurrent().parent.switch(0)

    s.resp_entry = None

  @blocking
  def amo( s, amo_type, addr, nbytes, data ):
    while s.req_entry is not None:
      greenlet.getcurrent().parent.switch(0)

    s.req_entry = s.create_req( amo_type, 0, addr, nbytes, data )

    while s.resp_entry is None:
      greenlet.getcurrent().parent.switch(0)

    ret = s.resp_entry.data[0:nbytes<<3]
    s.resp_entry = None
    return ret

  def construct( s, ReqType, RespType ):
    s.master = MasterIfcRTL( ReqType, RespType )

    # Use create_req to handle type mismatch
    Tlen  = ReqType.get_field_type('len')
    Tdata = ReqType.get_field_type('data')
    s.create_req = lambda a,b,c,d,e=0: ReqType( a, b, c, Tlen(d, trunc_int=True), Tdata(int(e)) )

    s.req_entry  = None
    s.resp_entry = None

    # req path

    s.req_sent = Wire()

    @update_ff
    def up_req_sent():
      s.req_sent <<= s.master.req.val & s.master.req.rdy

    @update
    def up_clear_req():
      if s.req_sent:
        s.req_entry = None

    @update_once
    def up_send_req():
      if s.req_entry is None:
        s.master.req.val @= 0
      else:
        s.master.req.val @= 1
        s.master.req.msg @= s.req_entry

    # resp path
    @update_once
    def up_resp_rdy():
      s.master.resp.rdy @= (s.resp_entry is None)

    @update_once
    def up_resp_msg():
      if (s.resp_entry is None) & s.master.resp.val:
        s.resp_entry = clone_deepcopy( s.master.resp.msg )

    s.add_constraints( U( up_clear_req ) < M(s.read),
                       U( up_clear_req ) < M(s.write),
                       U( up_clear_req ) < M(s.amo),

                       M( s.read )    < U( up_send_req ),
                       M( s.write )   < U( up_send_req ),
                       M( s.amo )     < U( up_send_req ),

                       M( s.read )      < U( up_resp_rdy ),
                       M( s.write )     < U( up_resp_rdy ),
                       M( s.amo )       < U( up_resp_rdy ),
                       U( up_resp_rdy ) < U( up_resp_msg ) )

class XcelMasterAdapter( Component ):

  @blocking
  def read( s, addr ):
    while s.req_entry is not None:
      greenlet.getcurrent().parent.switch(0)

    s.req_entry = s.create_req( 0, addr )

    while s.resp_entry is None:
      greenlet.getcurrent().parent.switch(0)

    ret = s.resp_entry.data
    s.resp_entry = None
    return ret

  @blocking
  def write( s, addr, data ):
    while s.req_entry is not None:
      greenlet.getcurrent().parent.switch(0)

    s.req_entry = s.create_req( 1, addr, data )

    while s.resp_entry is None:
      greenlet.getcurrent().parent.switch(0)

    s.resp_entry = None

  def construct( s, ReqType, RespType ):
    s.master = MasterIfcRTL( ReqType, RespType )

    # Use create_req to handle type mismatch
    Tdata = ReqType.get_field_type('data')
    s.create_req = lambda a,b,c=0: ReqType( a, b, Tdata(int(c)) )

    s.req_entry  = None
    s.resp_entry = None

    # req path

    s.req_sent = Wire()

    @update_ff
    def up_req_sent():
      s.req_sent <<= s.master.req.val & s.master.req.rdy

    @update
    def up_clear_req():
      if s.req_sent:
        s.req_entry = None

    @update_once
    def up_send_req():
      if s.req_entry is None:
        s.master.req.val @= 0
      else:
        s.master.req.val @= 1
        s.master.req.msg @= s.req_entry

    # resp path
    @update_once
    def up_resp_rdy():
      s.master.resp.rdy @= (s.resp_entry is None)

    @update_once
    def up_resp_msg():
      if (s.resp_entry is None) & s.master.resp.val:
        s.resp_entry = clone_deepcopy( s.master.resp.msg )

    s.add_constraints( U( up_clear_req ) < M(s.read),
                       U( up_clear_req ) < M(s.write),

                       M( s.read )    < U( up_send_req ),
                       M( s.write )   < U( up_send_req ),

                       M( s.read )      < U( up_resp_rdy ),
                       M( s.write )     < U( up_resp_rdy ),
                       U( up_resp_rdy ) < U( up_resp_msg ) )



#=========================================================================
# BehavioralRTLIRGenL5Pass.py
#=========================================================================
# Author : Peitian Pan
# Date   : Oct 20, 2018
"""Provide L5 behavioral RTLIR generation pass."""

from .BehavioralRTLIRGenL4Pass import (
    BehavioralRTLIRGeneratorL4,
    BehavioralRTLIRGenL4Pass,
)


class BehavioralRTLIRGenL5Pass( BehavioralRTLIRGenL4Pass ):
  # Override
  def get_rtlir_generator_class( s ):
    return BehavioralRTLIRGeneratorL5

class BehavioralRTLIRGeneratorL5( BehavioralRTLIRGeneratorL4 ):
  """Behavioral RTLIR generator level 5.

  Do nothing here because attributes have been handled in previous
  levels.
  """



#=========================================================================
# VImportSignalGen.py
#=========================================================================
# Author : Peitian Pan
# Date   : June 1, 2019
"""Test the SystemVerilog signal generation of imported component."""

from pymtl3.datatypes import Bits1, Bits32, bitstruct, mk_bits
from pymtl3.dsl import Component, InPort, Interface, OutPort
from pymtl3.passes.rtlir import RTLIRDataType as rdt
from pymtl3.passes.rtlir import RTLIRType as rt
from pymtl3.passes.rtlir.util.test_utility import do_test

from ..VerilogVerilatorImportPass import VerilogVerilatorImportPass


def local_do_test( m ):
  m.elaborate()
  rtype = rt.RTLIRGetter(cache=False).get_component_ifc_rtlir( m )
  ipass = VerilogVerilatorImportPass()
  symbols, decls = ipass.gen_signal_decl_py( rtype )
  assert symbols == m._ref_symbols
  assert decls == m._ref_decls

def test_port_single( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( mk_bits( 322 ) )
  a = A()
  a._ref_symbols = { 'Bits322' : mk_bits(322) }
  a._ref_decls = [
    "s.in_ = InPort( Bits322 )",
  ]
  do_test( a )

def test_port_array( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = [ InPort( Bits32 ) for _ in range( 3 ) ]
  a = A()
  a._ref_symbols = {}
  a._ref_decls = [
    "s.in_ = [ InPort( Bits32 ) for _ in range(3) ]",
  ]
  do_test( a )

def test_port_2d_array( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = [ [ InPort( Bits32 ) for _ in range(2) ] for _ in range(3) ]
  a = A()
  a._ref_symbols = {}
  a._ref_decls = [
    "s.in_ = [ [ InPort( Bits32 ) for _ in range(2) ] for _ in range(3) ]",
  ]
  do_test( a )

def test_struct_port_single( do_test ):
  @bitstruct
  class struct:
    bar: Bits32
    foo: Bits32
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( struct )
  a = A()
  a._ref_symbols = { 'struct__bar_32__foo_32' : struct }
  a._ref_decls = [
    "s.in_ = InPort( struct__bar_32__foo_32 )",
  ]
  do_test( a )

def test_struct_port_array( do_test ):
  @bitstruct
  class struct:
    bar: Bits32
    foo: Bits32
  class A( Component ):
    def construct( s ):
      s.in_ = [ InPort( struct ) for _ in range(2) ]
  a = A()
  a._ref_symbols = { 'struct__bar_32__foo_32' : struct }
  a._ref_decls = [
    "s.in_ = [ InPort( struct__bar_32__foo_32 ) for _ in range(2) ]",
  ]
  do_test( a )

def test_packed_array_port_array( do_test ):
  @bitstruct
  class struct:
    bar: Bits32
    foo: [ [ Bits32 for _ in range(2) ] for _ in range(3) ]
  class A( Component ):
    def construct( s ):
      s.in_ = [ InPort( struct ) for _ in range(2) ]
  a = A()
  a._ref_symbols = { 'struct__bar_32__foo_32x3x2' : struct }
  a._ref_decls = [
    "s.in_ = [ InPort( struct__bar_32__foo_32x3x2 ) for _ in range(2) ]",
  ]
  do_test( a )

def test_nested_struct( do_test ):
  @bitstruct
  class inner_struct:
    foo: Bits32
  @bitstruct
  class struct:
    bar: Bits32
    inner: inner_struct
  class A( Component ):
    def construct( s ):
      s.in_ = [ InPort( struct ) for _ in range(2) ]
  a = A()
  # Inner struct will not be added to `symbols` because struct
  # refers to it!
  a._ref_symbols = { 'struct__bar_32__inner_inner_struct__foo_32' : struct }
  a._ref_decls = [
    "s.in_ = [ InPort( struct__bar_32__inner_inner_struct__foo_32 ) for _ in range(2) ]",
  ]
  do_test( a )

def test_interface( do_test ):
  class Ifc( Interface ):
    def construct( s ):
      s.msg = InPort( Bits32 )
      s.val = InPort( Bits1 )
      s.rdy = OutPort( Bits1 )
  class A( Component ):
    def construct( s ):
      s.ifc = Ifc()
  a = A()
  a._ref_symbols = { 'Ifc' : Ifc }
  a._ref_decls = [
    "s.ifc = Ifc()"
  ]
  do_test( a )

def test_interface_parameter( do_test ):
  class Ifc( Interface ):
    def construct( s, Type, nbits_val, nbits_rdy ):
      s.msg = InPort( Type )
      s.val = InPort( mk_bits(nbits_val) )
      # Added support for BitsN values in case someone wants to do
      # tricky things like this.
      s.rdy = OutPort( mk_bits(nbits_rdy.nbits) )
  class A( Component ):
    def construct( s ):
      s.ifc = Ifc( Bits32, 1, Bits1(1) )
  a = A()
  a._ref_symbols = { 'Ifc' : Ifc }
  a._ref_decls = [
    "s.ifc = Ifc( Bits32, 1, Bits1( 1 ) )"
  ]
  do_test( a )

def test_interface_parameter_long_vector( do_test ):
  class Ifc( Interface ):
    def construct( s, Type, nbits_val, nbits_rdy ):
      s.msg = InPort( Type )
      s.val = InPort( mk_bits(nbits_val) )
      # Added support for BitsN values in case someone wants to do
      # tricky things like this.
      s.rdy = OutPort( mk_bits(nbits_rdy.nbits) )
  class A( Component ):
    def construct( s ):
      Bits322 = mk_bits(322)
      s.ifc = Ifc( Bits322, 1, Bits322(1) )
  a = A()
  a._ref_symbols = { 'Ifc' : Ifc, 'Bits322' : mk_bits(322) }
  a._ref_decls = [
    "s.ifc = Ifc( Bits322, 1, Bits322( 1 ) )"
  ]
  do_test( a )

def test_interface_array( do_test ):
  class Ifc( Interface ):
    def construct( s ):
      s.msg = InPort( Bits32 )
      s.val = InPort( Bits1 )
      s.rdy = OutPort( Bits1 )
  class A( Component ):
    def construct( s ):
      s.ifc = [ Ifc() for _ in range(2) ]
  a = A()
  a._ref_symbols = { 'Ifc' : Ifc }
  a._ref_decls = [
    "s.ifc = [ Ifc() for _ in range(2) ]"
  ]
  do_test( a )

def test_nested_interface( do_test ):
  class InnerIfc( Interface ):
    def construct( s ):
      s.msg = InPort( Bits32 )
      s.val = InPort( Bits1 )
      s.rdy = OutPort( Bits1 )
  class Ifc( Interface ):
    def construct( s ):
      s.valrdy_ifc = InnerIfc()
      s.ctrl_bar = InPort( Bits32 )
      s.ctrl_foo = OutPort( Bits32 )
  class A( Component ):
    def construct( s ):
      s.ifc = [ Ifc() for _ in range(2) ]
  a = A()
  # Inner interface will not be added to `symbols` because Ifc refers
  # to it!
  a._ref_symbols = { 'Ifc' : Ifc }
  a._ref_decls = [
    "s.ifc = [ Ifc() for _ in range(2) ]"
  ]
  do_test( a )

def test_nested_interface_port_array( do_test ):
  class InnerIfc( Interface ):
    def construct( s ):
      s.msg = [ InPort( Bits32 ) for _ in range(2) ]
      s.val = InPort( Bits1 )
      s.rdy = OutPort( Bits1 )
  class Ifc( Interface ):
    def construct( s ):
      s.valrdy_ifc = InnerIfc()
      s.ctrl_bar = InPort( Bits32 )
      s.ctrl_foo = OutPort( Bits32 )
  class A( Component ):
    def construct( s ):
      s.ifc = [ Ifc() for _ in range(2) ]
  a = A()
  # Inner interface will not be added to `symbols` because Ifc refers
  # to it!
  a._ref_symbols = { 'Ifc' : Ifc }
  a._ref_decls = [
    "s.ifc = [ Ifc() for _ in range(2) ]"
  ]
  do_test( a )



"""
========================================================================
ValRdyIfc
========================================================================
RTL val/rdy interface.

Author : Shunning Jiang
  Date : Apr 5, 2019
"""
from pymtl3 import *


def valrdy_to_str( msg, val, rdy, trace_len=15 ):
  if     val and not rdy: return "#".ljust( trace_len )
  if not val and     rdy: return " ".ljust( trace_len )
  if not val and not rdy: return ".".ljust( trace_len )
  return f"{msg}".ljust( trace_len ) # val and rdy

class RecvIfcRTL( Interface ):

  def construct( s, Type ):

    s.msg = InPort( Type )
    s.val = InPort()
    s.rdy = OutPort()

    s.trace_len = len(str(Type()))

  def __str__( s ):
    return valrdy_to_str( s.msg, s.val, s.rdy, s.trace_len )

class SendIfcRTL( Interface ):

  def construct( s, Type ):

    s.msg = OutPort( Type )
    s.val = OutPort()
    s.rdy = InPort()

    s.trace_len = len(str(Type()))

  def __str__( s ):
    return valrdy_to_str( s.msg, s.val, s.rdy, s.trace_len )

class MasterIfcRTL( Interface ):
  def construct( s, ReqType, RespType ):
    s.ReqType  = ReqType
    s.RespType = RespType
    s.req  = SendIfcRTL( Type=ReqType )
    s.resp = RecvIfcRTL( Type=RespType )
  def __str__( s ):
    return f"{s.req}|{s.resp}"

class MinionIfcRTL( Interface ):
  def construct( s, ReqType, RespType ):
    s.ReqType  = ReqType
    s.RespType = RespType
    s.req  = RecvIfcRTL( Type=ReqType )
    s.resp = SendIfcRTL( Type=RespType )
  def __str__( s ):
    return f"{s.req}|{s.resp}"



#=========================================================================
# BehavioralRTLIRGenL4Pass.py
#=========================================================================
# Author : Peitian Pan
# Date   : Oct 20, 2018
"""Provide L4 behavioral RTLIR generation pass."""

from .BehavioralRTLIRGenL3Pass import (
    BehavioralRTLIRGeneratorL3,
    BehavioralRTLIRGenL3Pass,
)


class BehavioralRTLIRGenL4Pass( BehavioralRTLIRGenL3Pass ):
  # Override
  def get_rtlir_generator_class( s ):
    return BehavioralRTLIRGeneratorL4

class BehavioralRTLIRGeneratorL4( BehavioralRTLIRGeneratorL3 ):
  """Behavioral RTLIR generator level 4.

  Do nothing here because attributes have been handled in previous
  levels.
  """



#=========================================================================
# VBehavioralTranslatorL1.py
#=========================================================================
# Author : Peitian Pan
# Date   : March 18, 2019
"""Provide the level 1 SystemVerilog translator implementation."""

from collections import deque
from contextlib import contextmanager

from pymtl3.datatypes import Bits, Bits32
from pymtl3.passes.backends.generic.behavioral.BehavioralTranslatorL1 import (
    BehavioralTranslatorL1,
)
from pymtl3.passes.rtlir import BehavioralRTLIR as bir
from pymtl3.passes.rtlir import RTLIRDataType as rdt
from pymtl3.passes.rtlir import RTLIRType as rt

from ...errors import VerilogTranslationError
from ...util.utility import make_indent
from .VBehavioralTranslatorL0 import VBehavioralTranslatorL0


class VBehavioralTranslatorL1( VBehavioralTranslatorL0, BehavioralTranslatorL1 ):

  def rtlir_tr_upblk_decls( s, upblk_decls ):
    ret = ''
    for idx, upblk_decl in enumerate(upblk_decls):
      make_indent( upblk_decl, 1 )
      if idx != 0:
        ret += '\n'
      ret += '\n' + '\n'.join( upblk_decl )
    return ret

  def rtlir_tr_upblk_decl( s, upblk, src, py_src ):
    return py_src + [ "" ] + src

  def _get_rtlir2v_visitor( s ):
    return BehavioralRTLIRToVVisitorL1

  def rtlir_tr_upblk_srcs( s, upblk_srcs ):
    ret = ''
    for upblk_src in upblk_srcs:
      make_indent( upblk_src, 1 )
      ret += '\n' + '\n'.join( upblk_src )
    return ret

  def rtlir_tr_upblk_src( s, upblk, rtlir_upblk ):
    visitor = s._get_rtlir2v_visitor()(s.is_verilog_reserved)
    return visitor.enter( upblk, rtlir_upblk )

  def rtlir_tr_upblk_py_srcs( s, upblk_py_srcs ):
    ret = ''
    for upblk_py_src in upblk_py_srcs:
      make_indent( upblk_py_src, 1 )
      ret += '\n' + '\n'.join( upblk_py_src )
    return ret

  def rtlir_tr_upblk_py_src( s, upblk, is_lambda, src, lino, filename ):
    def _trim( py_src ):
      indent = 100
      for line in py_src:
        if line:
          n_spaces = len( line ) - len( line.lstrip() )
          if n_spaces < indent:
            indent = n_spaces
      for idx, line in enumerate( py_src ):
        if line:
          py_src[ idx ] = line[indent:]

    upblk_py_src = src.split( '\n' )
    _trim( upblk_py_src )
    while upblk_py_src and not upblk_py_src[-1]:
      upblk_py_src = upblk_py_src[:-1]

    # Add comments to the generated block
    py_src = []

    if is_lambda:
      py_src += [ "PyMTL Lambda Block Source" ]
    else:
      py_src += [ "PyMTL Update Block Source" ]

    py_src += [ f"At {filename}:{lino}" ]

    py_src += upblk_py_src

    return ["// "+x for x in py_src]

  def rtlir_tr_behavioral_freevars( s, freevars ):
    make_indent( freevars, 1 )
    return '\n'.join( freevars )

  def rtlir_tr_behavioral_freevar( s, id_, rtype, array_type, dtype, obj ):
    assert isinstance( rtype, rt.Const ), \
      f'{id_} freevar should be a constant!'
    assert isinstance( rtype.get_dtype(), rdt.Vector ), \
      f'{id_} freevar should be a (list of) integer/BitStruct at L1!'
    return s.rtlir_tr_const_decl( '__const__'+id_, rtype, array_type, dtype, obj )

#-------------------------------------------------------------------------
# BehavioralRTLIRToVVisitorL1
#-------------------------------------------------------------------------

class BehavioralRTLIRToVVisitorL1( bir.BehavioralRTLIRNodeVisitor ):
  """Visitor that translates RTLIR to SystemVerilog for a single upblk."""

  def __init__( s, is_reserved ):
    # Should use enum here, but enum is a python 3 feature...
    s.NONE          = 0
    s.COMBINATIONAL = 1
    s.SEQUENTIAL    = 2
    s.upblk_type    = s.NONE
    s._is_verilog_reserved = is_reserved
    s._unpacked_q   = deque()

  def visit( s, node, *args ):
    # Customized epilogue processing
    method = 'visit_' + node.__class__.__name__
    visitor = getattr( s, method, s.generic_visit )
    ret = visitor( node, *args )
    return s.process_epilogue( node, ret )

  def process_epilogue( s, node, ret ):
    if hasattr( node, '_top_expr' ) and node._top_expr:
      # This node is a top level node -- we append all
      # pending indice.
      return s.process_unpacked_q( node, ret, f'{ret}{{}}')
    else:
      # This is a regular node. Just return its result.
      return ret

  def is_verilog_reserved( s, name ):
    return s._is_verilog_reserved( name )

  def process_unpacked_q( s, node, signal, signal_tplt ):
    if isinstance( node.Type, rt.Port ):
      filler = ''.join([f'[{i}]' for i in list(s._unpacked_q)])
      s._unpacked_q.clear()
      if '{}' in signal_tplt:
        p = signal_tplt.find('{}')
        ret = signal_tplt[:p] + filler + signal_tplt[p+2:]
      else:
        ret = signal_tplt
      return ret
    else:
      return signal

  def enter( s, blk, rtlir ):
    """Entry point for RTLIR generation."""
    s.blk     = blk

    # s.globals contains a dict of the global namespace of the module where
    # blk was defined
    s.globals = blk.__globals__

    # s.closure contains the free variables defined in an enclosing scope.
    # Basically this is the model instance s.
    s.closure = {}

    for i, var in enumerate( blk.__code__.co_freevars ):
      try:
        s.closure[ var ] = blk.__closure__[ i ].cell_contents
      except ValueError:
        pass

    # Are we inside the LHS of an assignemnt?

    s.is_assign_LHS = False

    return s.visit( rtlir )

  def check_res( s, node, name ):
    if s.is_verilog_reserved( name ):
      raise VerilogTranslationError( s.blk, node,
        f"name {name} is a SystemVerilog reserved keyword!" )

  #-----------------------------------------------------------------------
  # visit_CombUpblk
  #-----------------------------------------------------------------------

  def visit_CombUpblk( s, node ):
    """Return the Verilog representation of statements inside it."""
    blk_name = node.name
    src      = []
    body     = []
    s.upblk_type = s.COMBINATIONAL

    s.check_res( node, blk_name )

    # Add name of the upblk to this always block
    src.append( f'always_comb begin : {blk_name}' )

    for stmt in node.body:
      body.extend( s.visit( stmt ) )

    make_indent( body, 1 )
    src.extend( body )
    src.append( 'end' )
    s.upblk_type = s.NONE
    return src

  #-----------------------------------------------------------------------
  # visit_SeqUpblk
  #-----------------------------------------------------------------------

  def visit_SeqUpblk( s, node ):
    """Return the Verilog representation of statements inside it."""
    blk_name = node.name
    src      = []
    body     = []
    s.upblk_type = s.SEQUENTIAL

    s.check_res( node, blk_name )

    # Add name of the upblk to this always block
    src.append( f'always_ff @(posedge clk) begin : {blk_name}' )

    for stmt in node.body:
      body.extend( s.visit( stmt ) )

    make_indent( body, 1 )
    src.extend( body )
    src.append( 'end' )
    s.upblk_type = s.NONE
    return src

  #-----------------------------------------------------------------------
  # Statements
  #-----------------------------------------------------------------------
  # All statement nodes return a list of strings.

  #-----------------------------------------------------------------------
  # visit_Assign
  #-----------------------------------------------------------------------

  def visit_Assign( s, node ):
    for target in node.targets:
      target._top_expr = True
    node.value._top_expr = True

    with s.register_assign_LHS():
      targets     = [ s.visit( target ) for target in node.targets ]

    value         = s.visit( node.value )
    assignment_op = '<=' if not node.blocking else '='
    tplt = '{target} {assignment_op} {value};'

    return [ tplt.format(
      target = target, assignment_op = assignment_op, value = value
    ) for target in reversed(targets) ]

  # register_assign_LHS

  @contextmanager
  def register_assign_LHS( s ):
    assert not s.is_assign_LHS
    s.is_assign_LHS = True
    yield
    assert s.is_assign_LHS
    s.is_assign_LHS = False

  #-----------------------------------------------------------------------
  # visit_For
  #-----------------------------------------------------------------------

  def visit_For( s, node ):
    raise VerilogTranslationError( s.blk, node, "For not supported at L1" )

  #-----------------------------------------------------------------------
  # Expressions
  #-----------------------------------------------------------------------
  # All expression nodes return a single string.

  #-----------------------------------------------------------------------
  # visit_Number
  #-----------------------------------------------------------------------

  def visit_Number( s, node ):
    """Return a number in string."""
    nbits = node.Type.get_dtype().get_length()
    return f"{nbits}'d{node.value}"

  #-----------------------------------------------------------------------
  # visit_Concat
  #-----------------------------------------------------------------------

  def visit_Concat( s, node ):
    for value in node.values:
      value._top_expr = True

    values = [ s.visit(v) for v in node.values ]
    signals = ', '.join( values )
    return f'{{ {signals} }}'

  #-----------------------------------------------------------------------
  # visit_ZeroExt
  #-----------------------------------------------------------------------

  def visit_ZeroExt( s, node ):
    node.value._top_expr = True

    value = s.visit( node.value )
    target_nbits = node.nbits
    current_nbits = int(node.value.Type.get_dtype().get_length())
    padded_nbits = target_nbits - current_nbits
    if padded_nbits == 0:
      return value
    else:
      return f"{{ {{ {padded_nbits} {{ 1'b0 }} }}, {value} }}"

  #-----------------------------------------------------------------------
  # visit_SignExt
  #-----------------------------------------------------------------------
  # We need to special case the situation where a bit-selection or single-bit
  # part selection needs to be sign-extended ( Verilator throws an error
  # at this ) e.g. in_[31:31][0] should be translated into in_[31:31].

  def visit_SignExt( s, node ):
    node.value._top_expr = True

    value = s.visit( node.value )
    target_nbits = node.nbits
    current_nbits = int(node.value.Type.get_dtype().get_length())
    last_bit = current_nbits - 1
    padded_nbits = target_nbits - current_nbits

    if padded_nbits == 0:
      return value

    template = "{{ {{ {padded_nbits} {{ {value}[{last_bit}] }} }}, {value} }}"
    one_bit_template = "{{ {{ {padded_nbits} {{ {_value} }} }}, {value} }}"

    # Check if the signal to be extended is a bit selection or one-bit part
    # selection.
    if isinstance( node.value, bir.Slice ):
      try:
        lower = node.value.lower._value
        upper = node.value.upper._value
        if upper - lower == 1:
          _one_bit = True
        else:
          _one_bit = False
      except AttributeError:
        _one_bit = False

      # Manipulate the slicing string to avoid indexing on a sliced signal
      if not _one_bit:
        l, col, r = value.rfind('['), value.rfind(':'), value.rfind(']')
        if -1 < l < col < r:
          _value = value[:col] + ']'
          return one_bit_template.format( **locals() )

    elif isinstance( node.value, bir.Index ):
      _one_bit = True
    else:
      _one_bit = False

    if _one_bit:
      _value = value
      return one_bit_template.format( **locals() )
    else:
      return template.format( **locals() )

  #-----------------------------------------------------------------------
  # visit_Truncate
  #-----------------------------------------------------------------------
  # To quote the LRM about the size casting operator:
  # When changing the size, the cast shall return the value that a packed
  # array type with a single [n-1:0] dimension would hold after being
  # assigned the expression, where n is the cast size. The signedness shall
  # pass through unchanged.
  #
  # This means if we translate the truncation into a size casting expression,
  # the same semantics is preserved. This seems a better approach to me
  # than slicing because iirc slicing over integer literals is an error
  # in Verilator.

  def visit_Truncate( s, node ):
    nbits = node.nbits
    value = s.visit( node.value )
    dtype = node.value.Type.get_dtype()
    if isinstance(dtype, rdt.Vector) and dtype.get_length() > nbits:
      return f"{nbits}'({value})"
    else:
      return value

  #-----------------------------------------------------------------------
  # visit_Reduce
  #-----------------------------------------------------------------------

  def visit_Reduce( s, node ):
    node.value._top_expr = True

    op_t = type(node.op)
    reduce_ops = { bir.BitAnd : '&', bir.BitOr : '|', bir.BitXor : '^' }
    if op_t not in reduce_ops:
      raise VerilogTranslationError( s.blk, node,
          f"unrecognized operator {op_t} for reduce method!" )
    value = s.visit( node.value )
    op = reduce_ops[ op_t ]
    return f"( {op} {value} )"

  #-----------------------------------------------------------------------
  # visit_SizeCast
  #-----------------------------------------------------------------------

  def visit_SizeCast( s, node ):
    node.value._top_expr = True

    nbits = node.nbits
    value = s.visit( node.value )
    if hasattr( node, "_value" ):
      # value could be larger than what nbits can hold because it might
      # be a negative number. Since verilator errors when that happens, we
      # need to manually truncate the integer.
      value = int(Bits(nbits, node._value))
      return f"{nbits}'d{value}"

    return f"{nbits}'( {value} )"

  #-----------------------------------------------------------------------
  # visit_StructInst
  #-----------------------------------------------------------------------

  def visit_StructInst( s, node ):
    raise VerilogTranslationError( s.blk, node, "StructInst not supported at L1" )

  #-----------------------------------------------------------------------
  # visit_IfExp
  #-----------------------------------------------------------------------

  def visit_IfExp( s, node ):
    raise VerilogTranslationError( s.blk, node, "IfExp not supported at L1" )

  #-----------------------------------------------------------------------
  # visit_UnaryOp
  #-----------------------------------------------------------------------

  def visit_UnaryOp( s, node ):
    raise VerilogTranslationError( s.blk, node, "UnaryOp not supported at L1" )

  #-----------------------------------------------------------------------
  # visit_BoolOp
  #-----------------------------------------------------------------------

  def visit_BoolOp( s, node ):
    raise VerilogTranslationError( s.blk, node, "BoolOp not supported at L1" )

  #-----------------------------------------------------------------------
  # visit_BinOp
  #-----------------------------------------------------------------------

  def visit_BinOp( s, node ):
    raise VerilogTranslationError( s.blk, node, "BinOp not supported at L1" )

  #-----------------------------------------------------------------------
  # visit_Compare
  #-----------------------------------------------------------------------

  def visit_Compare( s, node ):
    raise VerilogTranslationError( s.blk, node, "Compare not supported at L1" )

  #-----------------------------------------------------------------------
  # visit_Attribute
  #-----------------------------------------------------------------------

  def visit_Attribute( s, node ):

    attr  = node.attr
    value = s.visit( node.value )

    s.check_res( node, attr )

    if isinstance( node.value, bir.Base ):
      # The base of this attribute node is the component 's'.
      # Example: s.out, s.STATE_IDLE
      # assert node.value.base is s.component
      if isinstance( node.Type, rt.Const ) and isinstance( node.Type.get_dtype(), rdt.Vector ):
        nbits = node.Type.get_dtype().get_length()
        ret = f"{nbits}'( {attr} )"
      else:
        ret = attr
    else:
      raise VerilogTranslationError( s.blk, node,
          "sub-components are not supported at L1" )

    return ret

  #-----------------------------------------------------------------------
  # visit_Index
  #-----------------------------------------------------------------------

  def visit_Index( s, node ):
    node.idx._top_expr = True

    idx   = s.visit( node.idx )
    value = s.visit( node.value )
    Type = node.value.Type

    # Unpacked index
    if isinstance( Type, rt.Array ):

      subtype = Type.get_sub_type()
      # Unpacked array index must be a static constant integer!
      # if idx is None and not isinstance(subtype, (rt.Port, rt.Wire, rt.Const)):
        # raise VerilogTranslationError( s.blk, node.ast,
# 'index of unpacked array {} must be a constant integer expression!'. \
            # format(node.value) )

      if isinstance( subtype, ( rt.Port, rt.Wire, rt.Const ) ):
        # Special case for situations like s.literal_int_array[0]
        # TODO: verify that these cases have been captured by the
        # constant extraction.
        assert not isinstance( node.Type, rt.Const )
        # if isinstance( node.Type, rt.Const ):
        #   nbits = node.Type.get_dtype().get_length()
        #   return f"{nbits}'({value}[{idx}])"

        return s.process_unpacked_q( node,
            f'{value}[{idx}]', f'{value}{{}}[{idx}]' )
      else:
        # is this branch ever taken?
        assert False
        return f'{value}__{idx}'

    # Index on a signal
    elif isinstance( Type, rt.Signal ):

      # Packed index
      if Type.is_packed_indexable():
        return f'{value}[{idx}]'
      # Bit selection
      elif isinstance( Type.get_dtype(), rdt.Vector ):
        return f'{value}[{idx}]'
      else:
        raise VerilogTranslationError( s.blk, node,
            "internal error: unrecognized index" )

    else:
      raise VerilogTranslationError( s.blk, node,
          "internal error: unrecognized index" )

  #-----------------------------------------------------------------------
  # visit_Slice
  #-----------------------------------------------------------------------

  def visit_Slice( s, node ):
    node.value._top_expr = True
    node.lower._top_expr = True
    node.upper._top_expr = True

    lower = s.visit( node.lower )
    value = s.visit( node.value )

    # Check for +: syntax
    if node.base and node.size:
      size = int( node.size )
      base = s.visit( node.base )

      return f'{value}[{base} +: {size}]'

    # Regular [ lower : upper ] syntax
    else:
      if hasattr( node.upper, '_value' ):
        upper = str( int( node.upper._value - 1 ) )
        nbits = node.upper.Type.get_dtype().get_length()
        upper = f"{nbits}'d{upper}"
      else:
        upper = s.visit( node.upper ) + '-1'

      return f'{value}[{upper}:{lower}]'

  #-----------------------------------------------------------------------
  # visit_Base
  #-----------------------------------------------------------------------

  def visit_Base( s, node ):
    return str( node.base )

  #-----------------------------------------------------------------------
  # visit_LoopVar
  #-----------------------------------------------------------------------

  def visit_LoopVar( s, node ):
    raise VerilogTranslationError( s.blk, node, "LoopVar not supported at L1" )

  #-----------------------------------------------------------------------
  # visit_FreeVar
  #-----------------------------------------------------------------------

  def visit_FreeVar( s, node ):
    nbits = node.Type.get_dtype().get_length()
    return f"{nbits}'( __const__{node.name} )"

  #-----------------------------------------------------------------------
  # visit_TmpVar
  #-----------------------------------------------------------------------

  def visit_TmpVar( s, node ):
    raise VerilogTranslationError( s.blk, node, "TmpVar not supported at L1" )

  #-----------------------------------------------------------------------
  # visit_LoopVarDecl
  #-----------------------------------------------------------------------

  def visit_LoopVarDecl( s, node ):
    raise VerilogTranslationError( s.blk, node, "LoopVarDecl not supported at L1" )



#=========================================================================
# VStructuralTranslatorL2.py
#=========================================================================
"""Provide SystemVerilog structural translator implementation."""

from textwrap import dedent

from pymtl3.passes.backends.generic.structural.StructuralTranslatorL2 import (
    StructuralTranslatorL2,
)
from pymtl3.passes.rtlir import RTLIRDataType as rdt

from ...util.utility import make_indent, pretty_concat
from .VStructuralTranslatorL1 import VStructuralTranslatorL1


class VStructuralTranslatorL2(
    VStructuralTranslatorL1, StructuralTranslatorL2 ):

  #-----------------------------------------------------------------------
  # Data types
  #-----------------------------------------------------------------------

  def rtlir_tr_packed_array_dtype( s, dtype ):
    sub_dtype = dtype.get_sub_dtype()
    if isinstance( sub_dtype, rdt.Vector ):
      sub_dtype_tr = s.rtlir_tr_vector_dtype( sub_dtype )
    elif isinstance( sub_dtype, rdt.Struct ):
      sub_dtype_tr = s.rtlir_tr_struct_dtype( sub_dtype )
    else:
      assert False, f"unsupported data type {sub_dtype} in packed array!"
    dim_str = "".join( f"[{size-1}:0]" for size in dtype.get_dim_sizes() )
    return {
      'def' : '',
      'data_type' : f"{sub_dtype_tr['data_type']}",
      'packed_type' : f"{dim_str}{sub_dtype_tr['packed_type']}",
      'unpacked_type' : sub_dtype_tr['unpacked_type'],
      'ndim' : dtype.get_dim_sizes(),
      'raw_dtype' : dtype
    }

  def rtlir_tr_struct_dtype( s, dtype ):
    dtype_name = dtype.get_name()
    field_decls = []

    for id_, _dtype in dtype.get_all_properties().items():

      if isinstance( _dtype, rdt.Vector ):
        tr = s.rtlir_tr_vector_dtype(_dtype)
      elif isinstance( _dtype, rdt.PackedArray ):
        tr = s.rtlir_tr_packed_array_dtype(_dtype)
      elif isinstance( _dtype, rdt.Struct ):
        tr = s.rtlir_tr_struct_dtype(_dtype)
      else:
        assert False, \
          f'unrecoganized field type {_dtype} of struct {dtype_name}!'
      field_decls.append(pretty_concat(tr['data_type'], tr['packed_type'], id_, ';'))

    make_indent( field_decls, 1 )
    field_decl = '\n'.join( field_decls )

    return {
      'def' : dedent("""\
        typedef struct packed {{
        {field_decl}
        }} {dtype_name};
        """).format(**locals()),
      'nbits' : dtype.get_length(),
      'data_type' : dtype_name,
      'packed_type' : '',
      'unpacked_type' : '',
      'raw_dtype' : dtype
    }

  #-----------------------------------------------------------------------
  # Declarations
  #-----------------------------------------------------------------------

  def gen_array_param( s, n_dim, dtype, array ):
    if not n_dim and isinstance( dtype, rdt.Struct ):
      return s.rtlir_tr_struct_instance( dtype, array )
    else:
      return super().gen_array_param( n_dim, dtype, array )

  #-----------------------------------------------------------------------
  # Signal oeprations
  #-----------------------------------------------------------------------

  def rtlir_tr_packed_index( s, base_signal, index, status ):
    return s._rtlir_tr_process_unpacked(
              f'{base_signal}[{index}]',
              f'{base_signal}{{}}[{index}]',
              status, ('status', 'unpacked') )

  def rtlir_tr_struct_attr( s, base_signal, attr, status ):
    return s._rtlir_tr_process_unpacked(
              f'{base_signal}.{attr}',
              f'{base_signal}{{}}.{attr}',
              status, ('status', 'unpacked') )

  def rtlir_tr_struct_instance( s, dtype, struct ):
    def _gen_packed_array( dtype, n_dim, array ):
      if not n_dim:
        if isinstance( dtype, rdt.Vector ):
          return s.rtlir_tr_literal_number( dtype.nbits, array )
        elif isinstance( dtype, rdt.Struct ):
          return s.rtlir_tr_struct_instance( dtype, array )
        else:
          assert False, f"unrecognized data type {dtype}!"
      else:
        ret = []
        for i in reversed( range( n_dim[0]) ):
          ret.append( _gen_packed_array( dtype, n_dim[1:], array[i] ) )
        if n_dim[0] > 1:
          cat_str = "{" + ", ".join( ret ) + "}"
        else:
          cat_str = ", ".join( ret )
        return f"{{ {cat_str} }}"
    ret = []
    for name, Type in dtype.get_all_properties().items():
      field = getattr( struct, name )
      if isinstance( Type, rdt.Vector ):
        _ret = s.rtlir_tr_literal_number( Type.nbits, field )
      elif isinstance( Type, rdt.Struct ):
        _ret = s.rtlir_tr_struct_instance( Type, field )
      elif isinstance( Type, rdt.PackedArray ):
        n_dim = Type.get_dim_sizes()
        sub_dtype = Type.get_sub_dtype()
        _ret = _gen_packed_array( sub_dtype, n_dim, field )
      else:
        assert False, f"unrecognized data type {Type}!"
      ret.append( _ret )
    return f"{{ {', '.join(ret)} }}"
